/*
 * Copyright (c) 2007 Hypertriton, Inc. <http://hypertriton.com/>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <config/version.h>
#include <config/release.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>

#include <string.h>
#include <errno.h>
#include <syslog.h>
#include <pwd.h>
#include <stdio.h>
#include <unistd.h>

#include "mailprocd.h"
#include "pathnames.h"
#include "peercred.h"

char *ctlSockPath = _PATH_CTL_SOCKET;	/* Path to control socket */
unsigned int ctlSessionMax = 50;	/* Maximum concurrent sessions */

struct ctl_session_q ctlSessions;
unsigned int ctlSessionCount;

const char *ctlCommandNames[] = {
	"CTL_VERSION",
	"CTL_VERSION_SA",
	"CTL_SPAMCHECK_EXIT",
	NULL
};

int
CTL_Init(void)
{
	TAILQ_INIT(&ctlSessions);
	ctlSessionCount = 0;
	return (0);
}

void
CTL_Destroy(void)
{
	CTL_Session *sess, *sessNext;

	for (sess = TAILQ_FIRST(&ctlSessions);
	     sess != TAILQ_END(&ctlSessions);
	     sess = sessNext) {
		sessNext = TAILQ_NEXT(sess, sessions);
		free(sess);
	}
	TAILQ_INIT(&ctlSessions);
	unlink(ctlSockPath);
}

void
CTL_ForkInstance(int sock)
{
	pid_t pid;

//	if (ctlSessionCount+1 > ctlSessionMax) {
//		syslog(LOG_ERR, "Too many control sessions (%d)",
//		    ctlSessionMax);
//		return;
//	}
	if ((pid = fork()) != 0) {
		CTL_Session *sess;

		sess = Malloc(sizeof(CTL_Session));
		sess->pid = pid;
		TAILQ_INSERT_TAIL(&ctlSessions, sess, sessions);
		Debug("New control session: %d (session #%d/%d)", (int)pid,
		    ctlSessionCount, ctlSessionMax);
		ctlSessionCount++;
	} else {
		dup2(sock, STDIN_FILENO);
		dup2(sock, STDOUT_FILENO);

		SF_EnterServerProc("mailprocd");
		Setproctitle("control");
		setvbuf(stdout, NULL, _IONBF, 0);
		CTL_Main();
		shutdown(fileno(stdin), SHUT_RDWR);
		shutdown(fileno(stdout), SHUT_RDWR);
		SF_ExitServerProc(0);
	}
}

static void
ServerResponse(int rv, const char *errMsg)
{
	char code = (rv == 0) ? '0' : '!';
	char errBuf[CTL_ERROR_MAX];

	write(STDOUT_FILENO, &code, 1);
	if (rv != 0) {
		memset(errBuf, '\0', sizeof(errBuf));
		Strlcpy(errBuf, errMsg, sizeof(errBuf));
		write(STDOUT_FILENO, errBuf, sizeof(errBuf));
	}
}

static int
DoSpamcheckExit(void)
{
	char path[1024];
	struct passwd *pw;
	uid_t uid;
	gid_t gid;

	if (Getpeereid(STDIN_FILENO, &uid, &gid) == -1) {
		return (-1);
	}
	Debug("SPAMCHECK_EXIT (uid = %d, gid = %d)",
	    (int)uid, (int)gid);
	if ((pw = getpwuid(uid)) == NULL) {
		SF_SetError("getpwuid");
		return (-1);
	}
	Strlcpy(path, _PATH_SOCKETDIR, sizeof(path));
	Strlcat(path, pw->pw_name, sizeof(path));
	if (unlink(path) == -1) {
		SF_SetError("%s: %s", path, strerror(errno));
		return (-1);
	}
	return (0);
}

void
CTL_Main(void)
{
	char cmd;

	if (write(STDOUT_FILENO, "mailprocd\n", 10) != 10) {
		goto fail_write;
	}
	if (read(STDIN_FILENO, &cmd, 1) != 1) {
		goto fail_read;
	}
	switch (cmd) {
	case CTL_SPAMCHECK_EXIT:
		if (DoSpamcheckExit() == 0) {
			ServerResponse(0, NULL);
		} else {
			ServerResponse(-1, SF_GetError());
			syslog(LOG_ERR, "CTL_SPAMCHECK_EXIT: %s",
			    SF_GetError());
		}
		break;
	case CTL_VERSION:
		write(STDOUT_FILENO, VERSION, sizeof(VERSION));
		write(STDOUT_FILENO, "\n", 1);
		write(STDOUT_FILENO, RELEASE, sizeof(RELEASE));
		write(STDOUT_FILENO, "\n", 1);
		break;
	case CTL_VERSION_SA:
#ifdef HAVE_SA
		write(STDOUT_FILENO, saVersion, strlen(saVersion));
		write(STDOUT_FILENO, "\n", 1);
#else
		ServerResponse(-1, "Not available");
#endif
		break;
	default:
		ServerResponse(-1, "Unimplemented command");
		Debug("Bad command: 0x%x", cmd);
		syslog(LOG_ERR, "Invalid command 0x%x on control socket",
		    cmd);
		break;
	}
	return;
fail_write:
	syslog(LOG_ERR, "Write error on control socket");
	return;
fail_read:
	syslog(LOG_ERR, "Read error on control socket");
	return;
}

/* Issue a request on the control socket. */
int
CTL_SendCommand(enum ctl_command cmd, ...)
{
	char signature[11];
	char errMsg[CTL_ERROR_MAX];
	struct sockaddr_un sun;
	my_socklen_t socklen;
	unsigned char cmdData, errCode;
	int s;

	Debug("Sending %s", ctlCommandNames[cmd]);
	if ((s = socket(AF_UNIX, SOCK_STREAM, 0)) == -1) {
		SF_SetError("socket: %s", strerror(errno));
		return (-1);
	}
	sun.sun_family = AF_UNIX;
	Strlcpy(sun.sun_path, _PATH_CTL_SOCKET, sizeof(sun.sun_path));
	socklen = SUN_LEN(&sun);
	if (connect(s, (struct sockaddr *)&sun, socklen) == -1) {
		SF_SetError("%s: %s", _PATH_CTL_SOCKET, strerror(errno));
		return (-1);
	}
	if (read(s, signature, 10) != 10) {
		SF_SetError("Error reading signature");
		goto fail;
	}
	signature[10] = '\0';
	if (strcmp(signature, "mailprocd\n") != 0) {
		SF_SetError("Bad signature (`%s' != `%s')", signature,
		    "mailprocd\n");
		goto fail;
	}
	cmdData = (unsigned char)cmd;
	if (write(s, &cmdData, 1) != 1) {
		SF_SetError("Error writing %s", ctlCommandNames[cmd]);
		goto fail;
	}
	if (read(s, &errCode, 1) != 1) {
		SF_SetError("Error reading response code");
		goto fail;
	}
	if (errCode == '!') {
		if (read(s, errMsg, sizeof(errMsg)) != sizeof(errMsg)) {
			SF_SetError("Error reading error message");
			goto fail;
		}
		errMsg[CTL_ERROR_MAX-1] = '\0';	
		SF_SetError("Server: %s", errMsg);
		goto fail;
	}
	close(s);
	Debug("Sent %s command", ctlCommandNames[cmd]);
	return (0);
fail:
	close(s);
	return (-1);
}
