/*
 * Copyright (c) 2006-2007 Hypertriton, Inc. <http://www.hypertriton.com/>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>

#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <stdarg.h>
#include <errno.h>
#include <unistd.h>
#include <pwd.h>
#include <grp.h>
#include <ctype.h>
#include <netdb.h>

#include "mailprocd.h"

/* RFC 1652 */
/* #define _8BITMIME */

int smtpMasterPipe = -1;		/* Pipe to master process */
int smtpMsgsTot;			/* Total messages processed */

static int looping = 0;			/* Loop detected during DATA */
static int outofmem = 0;		/* Out of memory during DATA */
static int inHeader = 1;		/* Currently reading header */
static int inData = 0;			/* DATA phase */
static SF_Message *msg = NULL;
static enum {
	ST_HELO,
	ST_MAIL_FROM,
	ST_RCPT_TO,
	ST_DATA,
	ST_DATA_IN,
	ST_QUIT
} state = ST_HELO;

static __inline__ void
FlushClient(void)
{
	fflush(stdout);
}

static __inline__ void
WriteEHLO(void)
{
	fputs("250-localhost\n"
	      "250-PIPELINING\n"
#ifdef _8BITMIME
	      "250-8BITMIME\n"
#endif
              "250-SIZE 104857600\n"
              "250 ENHANCEDSTATUSCODES\n", stdout);
	FlushClient();
}

static __inline__ void
WriteLHLO(void)
{
	fputs("250-localhost\n"
	      "250-PIPELINING\n"
#ifdef _8BITMIME
	      "250-8BITMIME\n"
#endif
              "250-SIZE 104857600\n"
              "250 ENHANCEDSTATUSCODES\n", stdout);
	FlushClient();
}

static __inline__ void
WriteClient(int doFlush, const char *code, const char *ext, const char *s)
{
	fputs(code, stdout);
	fputc(' ', stdout);
	if (ext != NULL) {
		fputs(ext, stdout);
		fputc(' ', stdout);
	}
	fputs(s, stdout);
	fputc('\n', stdout);

//	if (doFlush)
		FlushClient();
}

static int
Expect(FILE *f, const char *code)
{
	char *buf, *lbuf = NULL;
	size_t len;
	int rv = 0;

	if ((buf = Fgetln(f, &len)) == NULL) {
		SF_SetError("EOF from server");
		return (0);
	}
	if (buf[len-1] == '\n') {
		buf[len-1] = '\0';
	} else {
		if ((lbuf = malloc(len+1)) == NULL) {
			SF_SetError("malloc linebuf");
			return (0);
		}
		memcpy(lbuf, buf, len);
		lbuf[len] = '\0';
		buf = lbuf;
	}
	if (strncmp(buf, code, 3) != 0) {
		SF_SetError("returned \"%s\" (was expecting %s)", buf, code);
		rv = 0;
	}
	Free(lbuf);
	return (1);
}

/*
 * Forward a message back to the MTA, or place it in the queue if the
 * MTA is not responding.
 */
int
SMTP_ForwardMessage(SF_Message *msg, const char *rhost, const char *rport,
    const char *rcpt)
{
	struct addrinfo hints, *res, *res0;
	const char *cause = NULL;
	int rv, s;
	FILE *f;

	Debug("Forwarding message to <%s> via [%s:%s] (%lu bytes)", 
	    rcpt, rhost, rport, (unsigned long)msg->text_len);

	memset(&hints, 0, sizeof(hints));
	hints.ai_family = PF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	if ((rv = getaddrinfo(rhost, rport, &hints, &res0)) != 0) {
		SF_SetError("%s:%s: %s", rhost, rport, gai_strerror(rv));
		return (-1);
	}
	for (s = -1, res = res0;
	     res != NULL;
	     res = res->ai_next) {
		s = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
		if (s < 0) {
			cause = "socket";
			continue;
		}
		if (connect(s, res->ai_addr, res->ai_addrlen) < 0) {
			cause = "connect";
			close(s);
			s = -1;
			continue;
		}
		break;
	}
	if (s == -1) {
		SF_SetError("%s: %s", cause, strerror(errno));
		goto fail_addr;
	}
	if ((f = fdopen(s, "w+")) == NULL) {
		SF_SetError("fdopen: %s", strerror(errno));
		goto fail;
	}
	if (setvbuf(f, NULL, _IOLBF, 0) == EOF) {
		syslog(LOG_ERR, "setvbuf _IOLBF failed");
	}

	if (!Expect(f, "220")) { goto fail; }
	fputs("HELO [127.0.0.1]\n", f);
	if (!Expect(f, "250")) { goto fail; }
	fputs("MAIL FROM:<", f);
	fputs(msg->mail_from, f);
	fputs(">\n", f);
	if (!Expect(f, "250")) { goto fail; }
	fputs("RCPT TO:<", f);
	fputs(rcpt, f);
	fputs(">\n", f);
	if (!Expect(f, "250")) { goto fail; }
	fputs("DATA\n", f);
	if (!Expect(f, "354")) { goto fail; }
	fputs(msg->text, f);
	fputs("\r\n.\r\n", f);
	if (!Expect(f, "250")) { goto fail; }

	Debug("Reinjected message successfully");

	close(s);
	freeaddrinfo(res0);
	return (0);
fail:
	close(s);
fail_addr:
	freeaddrinfo(res0);
	return (-1);
}

/*
 * Remove whitespace/brackets from an address and convert to lowercase.
 * Destination must be able to hold ADDRESS_MAX characters.
 */
static int
GetAddress(char *dest, char *buf)
{
	char *c, *cStart;

	for (c = &buf[0]; *c == ' ' || *c == '<'; c++)
		;;
	cStart = c;

	for (; *c != '>' && *c != '\0'; c++)
		;;
	if (*c == '>')
		*c = '\0';

	if (Strlcpy(dest, cStart, ADDRESS_MAX) >= ADDRESS_MAX) {
		return (-1);
	}
	for (c = &dest[0]; *c != '\0'; c++) {
		*c = tolower((int)*c);
	}
	return (0);
}

static void
GoodBye(void)
{
	WriteClient(1, "221", "2.0.0", "Bye");
	shutdown(fileno(stdin), SHUT_RDWR);
	shutdown(fileno(stdout), SHUT_RDWR);
}

/* Respond to a bad command or one issued out of sequence. */
static void
InvalidCmd(const char *expected, const char *cmd)
{
	if (expected != NULL && cmd != NULL) {
		syslog(LOG_ERR, "SMTP Error: Expected %s, got `%s'",
		    expected, cmd);
	}
	WriteClient(1, "503", "5.5.1", "Invalid command");
	fflush(stdout);
}

static __inline__ void
ResetState(void)
{
	if (msg != NULL) {
		SF_MessageFree(msg);
		msg = NULL;
	}
	looping = 0;
	outofmem = 0;
}

static void
ProcessMailLoop(SF_Message *msg)
{
	SF_Recipient *rcpt;

	ResetState();
	WriteClient(1, "550", "5.4.6", "User mail-to-command loop detected!");
	/* XXX */
	TAILQ_FOREACH(rcpt, &msg->rcpts, rcpts) {
		syslog(LOG_ERR, "Mail loop with <%s>?", rcpt->addr);
		Debug("Mail loop detected under <%s>?", rcpt->addr);
	}
}

/* Process an incoming SMTP/LMTP session initiated by the MTA. */
int
SMTP_Main(enum smtp_prot prot)
{
	char *buf, *lbuf = NULL, *pbuf;
	size_t len;

	WriteClient(1, "220", NULL, (prot == ESMTP_PROTOCOL) ?
	    "localhost ESMTP" :
	    "localhost LMTP");

	while ((buf = Fgetln(stdin, &len)) != NULL) {
		if (buf[len-1] == '\n') {
			if (len > 1 && buf[len-2] == '\r') {
				len--;
			}
			buf[len-1] = '\0';
		} else {
			if ((lbuf = malloc(len+1)) == NULL) {
				SF_SetError("malloc line");
				goto fail;
			}
			memcpy(lbuf, buf, len);
			lbuf[len] = '\0';
			buf = lbuf;
		}

		if (!inData) {
			/*
			 * Commands allowed in any context other than DATA.
			 */
			if (strcasecmp(buf, "quit") == 0) {
				GoodBye();
				goto out;
			}
			if (strncasecmp(buf, "rset", 4) == 0) {
				ResetState();
				WriteClient(0, "250", "2.0.0", "Ok");
				state = ST_MAIL_FROM;
				Setproctitle("-RSET (%d msgs)", smtpMsgsTot);
				goto next_line;
			}
			if (strncasecmp(buf, "noop", 4) == 0) {
				WriteClient(1, "250", "2.0.0", "Ok");
				goto next_line;
			}
			if (state != ST_HELO &&
			    (strncasecmp(buf, "helo", 4) == 0 ||
			     strncasecmp(buf, "ehlo", 4) == 0 ||
			     strncasecmp(buf, "lhlo", 4) == 0)) {
				/*
				 * Per RFC2821, we reset the state as
				 * if RSET had been issued.
				 */
				state = ST_HELO;
			}
		}
reprocess:
		switch (state) {
		case ST_HELO:
			if (strncasecmp(buf, "helo", 4) == 0) {
				Setproctitle("HELO");
				WriteClient(1, "221", NULL, "localhost");
				state = ST_MAIL_FROM;
			} else if (strncasecmp(buf, "ehlo", 4) == 0) {
				WriteEHLO();
				state = ST_MAIL_FROM;
			} else if (strncasecmp(buf, "lhlo", 4) == 0) {
				WriteLHLO();
				state = ST_MAIL_FROM;
			} else if (strncasecmp(buf, "mail from", 9) == 0) {
				state = ST_MAIL_FROM;
				goto reprocess;
			} else {
				InvalidCmd("HELO/EHLO", buf);
			}
			break;
		case ST_MAIL_FROM:
			if (strncasecmp(buf, "mail from:", 10) == 0) {
				Setproctitle("-MAIL FROM");
				ResetState();
				if ((msg = SF_MessageNew()) == NULL) {
					WriteClient(1, "452", "4.3.2",
					               "Out of memory");
					goto fail;
				}
				if ((msg->text = malloc(1)) == NULL) {
					WriteClient(1, "452", "4.3.2",
					               "Out of memory");
					SF_SetError("Out of memory");
					goto fail;
				}
				msg->text[0] = '\0';
				msg->text_len = 0;
				if (GetAddress(msg->mail_from, &buf[10]) == -1){
					WriteClient(1, "503", "5.1.7",
					            "Bad sender address");
					ResetState();
					state = ST_HELO;
					break;
				}
				WriteClient(0, "250", "2.1.0", "Ok");
				state = ST_RCPT_TO;
			} else {
				InvalidCmd("MAIL FROM", buf);
			}
			break;
		case ST_RCPT_TO:
			if (strncasecmp(buf, "rcpt to:", 8) == 0) {
				SF_Recipient *rcpt;

				Setproctitle("-RCPT TO");
				if (msg == NULL) {
					InvalidCmd("FROM", buf);
					break;
				}
				if ((rcpt = SF_MessageAddRecipient(msg, NULL))
				    == NULL) {
					WriteClient(1, "452", "4.3.2",
					               "Out of memory");
					SF_SetError("Out of memory");
					goto fail;
				}
				if (GetAddress(rcpt->addr, &buf[8]) == -1 ||
				    SF_ParseRecipientParts(rcpt) == -1) {
					WriteClient(1, "503", "5.1.3",
					            "Bad recipient address");
					ResetState();
					state = ST_HELO;
					break;
				}
				WriteClient(0, "250", "2.1.5", "Ok");
				state = ST_DATA;
			} else {
				InvalidCmd("RCPT TO", buf);
			}
			break;
		case ST_DATA:
			if (msg == NULL) {
				WriteClient(1, "503", "5.5.1",
				               "DATA without RCPT/TO");
				break;
			}
			if (strncasecmp(buf, "data", 4) == 0) {
				Setproctitle("-DATA: %d msgs", smtpMsgsTot);
				WriteClient(1, "354", NULL,
				               "End data with "
					       "<CR><LF>.<CR><LF>");
				state = ST_DATA_IN;
				inData = 1;
				inHeader = 1;
			} else if (strncasecmp(buf, "mail from:", 10) == 0) {
				state = ST_MAIL_FROM;
				goto reprocess;
			} else if (strncasecmp(buf, "rcpt to:", 8) == 0) {
				state = ST_RCPT_TO;
				goto reprocess;
			} else {
				InvalidCmd("DATA/RCPT TO", buf);
			}
			break;
		case ST_DATA_IN:
			if (msg == NULL) {
				WriteClient(1, "503", "5.5.1",
				               "DATA without RCPT/TO");
				break;
			}
			if (buf[0] == '.' && buf[1] == '\0') {
				SF_Recipient *rcpt;
				int nsuccess = 0;

				inHeader = 0;
				inData = 0;
				if (outofmem) {
					WriteClient(1, "452", "4.3.2",
					               "Out of memory");
					SF_SetError("Out of memory");
					goto fail;
				}
				if (looping) {
					ProcessMailLoop(msg);
					Setproctitle("-QUIT: %d msgs",
					    smtpMsgsTot++);
					state = ST_QUIT;
					break;
				}
				msg->text[msg->text_len] = '\0';

				TAILQ_FOREACH(rcpt, &msg->rcpts, rcpts) {
					if (QMGR_Queue(msg, rcpt,
					    smtpMasterPipe) == 0) {
						nsuccess++;
					} else {
						syslog(LOG_ERR,
						    "QMGR Error: %s",
						    SF_GetError());
						Debug("QMGR: %s\n",
						    SF_GetError());
					}
				}
				if (nsuccess > 0) {
					WriteClient(1, "250", "2.0.0", "Ok");
				} else {
					WriteClient(1, "452", "4.3.1",
					               "Mail system full");
				}
				Setproctitle("-QUIT: %d msgs", smtpMsgsTot++);
				ResetState();
				state = ST_QUIT;
				break;
			}
			if (outofmem || looping) {
				break;
			}
			if (buf[0] == '\0') {
				inHeader = 0;
			} else if (!inHeader && buf[0] == 'F' &&
			    strncmp(buf, "From", 4) == 0) {
				/*
				 * Quote /^From/ lines for future
				 * mailbox delivery (it is more
				 * efficient to do it here).
				 */
				pbuf = realloc(msg->text, msg->text_len+1);
				if (pbuf == NULL) {
					outofmem = 1;
					break;
				}
				msg->text = pbuf;
				msg->text[msg->text_len] = '>';
				msg->text_len++;
			} else if (inHeader && buf[0] == 'X' &&
			    strncmp(buf, "X-Csoft-Pipe: ", 14) == 0) {
				SF_Recipient *xRcpt;

				TAILQ_FOREACH(xRcpt, &msg->rcpts, rcpts) {
					if (strcasecmp(&buf[14], xRcpt->addr)
					    == 0) {
						looping = 1;
						break;
					}
				}
			}
			pbuf = realloc(msg->text, msg->text_len+len+1);
			if (pbuf == NULL) {
				outofmem = 1;
				break;
			}
			msg->text = pbuf;
			buf[len-1] = '\n';
			memcpy(&msg->text[msg->text_len], buf, len);
			msg->text_len += len;
			break;
		case ST_QUIT:
			if (strncasecmp(buf, "quit", 4) == 0) {
				GoodBye();
				goto out;
			} else if (strncasecmp(buf, "mail from:", 10) == 0) {
				state = ST_MAIL_FROM;
				goto reprocess;
			} else if (strncasecmp(buf, "rcpt to:", 8) == 0) {
				state = ST_RCPT_TO;
				goto reprocess;
			} else {
				InvalidCmd(NULL, NULL);
				break;
			}
			break;
		default:
			Debug("SMTP invalid command: %s\n", buf);
			InvalidCmd(NULL, NULL);
			break;
		}
next_line:
		Free(lbuf);
		lbuf = NULL;
	}
	if (ferror(stdin)) {
		SF_SetError("Read error");
		goto fail;
	}
out:
	Free(lbuf);
	if (msg != NULL) { SF_MessageFree(msg); }
	return (0);
fail:
	Free(lbuf);
	if (msg != NULL) { SF_MessageFree(msg); }
	return (-1);
}

